@@ -77,13 +77,15 @@ class Stream extends AbstractStream implements HeaderEmitter {
     // State machine would be too much overhead
     private int headerState = HEADER_STATE_START;
     private StreamException headerException = null;
-    // TODO: null these when finished to reduce memory used by closed stream
-    private final Request coyoteRequest;
-    private StringBuilder cookieHeader = null;
-    private final Response coyoteResponse = new Response();
-    private final StreamInputBuffer inputBuffer;
-    private final StreamOutputBuffer streamOutputBuffer = new StreamOutputBuffer();
-    private final Http2OutputBuffer http2OutputBuffer =
+
+    // These will be set to null once the Stream closes to reduce the memory
+    // footprint.
+    private volatile Request coyoteRequest;
+    private volatile StringBuilder cookieHeader = null;
+    private volatile Response coyoteResponse = new Response();
+    private volatile StreamInputBuffer inputBuffer;
+    private volatile StreamOutputBuffer streamOutputBuffer = new StreamOutputBuffer();
+    private volatile Http2OutputBuffer http2OutputBuffer =
             new Http2OutputBuffer(coyoteResponse, streamOutputBuffer);
 
 
@@ -273,7 +275,7 @@ class Stream extends AbstractStream implements HeaderEmitter {
                     allocationManager.waitForStream(writeTimeout);
                     windowSize = getWindowSize();
                     if (windowSize == 0) {
-                        doWriteTimeout();
+                        doStreamCancel(sm.getString("stream.writeTimeout"), Http2Error.ENHANCE_YOUR_CALM);
                     }
                 } catch (InterruptedException e) {
                     // Possible shutdown / rst or similar. Use an IOException to
@@ -297,10 +299,8 @@ class Stream extends AbstractStream implements HeaderEmitter {
     }
 
 
-    void doWriteTimeout() throws CloseNowException {
-        String msg = sm.getString("stream.writeTimeout");
-        StreamException se = new StreamException(
-                msg, Http2Error.ENHANCE_YOUR_CALM, getIdAsInt());
+    void doStreamCancel(String msg, Http2Error error) throws CloseNowException {
+        StreamException se = new StreamException(msg, error, getIdAsInt());
         // Prevent the application making further writes
         streamOutputBuffer.closed = true;
         // Prevent Tomcat's error handling trying to write
@@ -583,6 +583,11 @@ class Stream extends AbstractStream implements HeaderEmitter {
 
 
     final ByteBuffer getInputByteBuffer() {
+        // Avoid NPE if Stream has been closed on Stream specific thread
+        StreamInputBuffer inputBuffer = this.inputBuffer;
+        if (inputBuffer == null) {
+            return null;
+        }
         return inputBuffer.getInBuffer();
     }
 
@@ -613,6 +618,11 @@ class Stream extends AbstractStream implements HeaderEmitter {
 
     final void receivedData(int payloadSize) throws ConnectionException {
         contentLengthReceived += payloadSize;
+        Request coyoteRequest = this.coyoteRequest;
+        // Avoid NPE if Stream has been closed on Stream specific thread
+        if (coyoteRequest == null) {
+            return;
+        }
         long contentLengthHeader = coyoteRequest.getContentLengthLong();
         if (contentLengthHeader > -1 && contentLengthReceived > contentLengthHeader) {
             throw new ConnectionException(sm.getString("stream.header.contentLength",
@@ -715,6 +725,7 @@ class Stream extends AbstractStream implements HeaderEmitter {
                             se.getError()));
                 }
                 state.sendReset();
+                cancelAllocationRequests();
                 handler.sendStreamReset(se);
             } catch (IOException ioe) {
                 ConnectionException ce = new ConnectionException(
@@ -725,6 +736,30 @@ class Stream extends AbstractStream implements HeaderEmitter {
         } else {
             handler.closeConnection(http2Exception);
         }
+        recycle();
+    }
+
+
+    /*
+     * This method is called recycle for consistency with the rest of the Tomcat
+     * code base. Currently, it only sets references to null for the purposes of
+     * reducing memory footprint. It does not fully recycle the Stream ready for
+     * re-use since Stream objects are not re-used. This is useful because
+     * Stream instances are retained for a period after the Stream closes.
+     */
+    final void recycle() {
+        if (log.isDebugEnabled()) {
+            log.debug(sm.getString("stream.recycle", getConnectionId(), getIdentifier()));
+        }
+        /*
+         * Temporarily disabled due to multiple regressions (NPEs)
+        coyoteRequest = null;
+        cookieHeader = null;
+        coyoteResponse = null;
+        inputBuffer = null;
+        streamOutputBuffer = null;
+        http2OutputBuffer = null;
+        */
     }
 
 
@@ -1151,7 +1186,8 @@ class Stream extends AbstractStream implements HeaderEmitter {
         }
 
 
-        final synchronized int available() {
+        @Override
+        public final synchronized int available() {
             if (inBuffer == null) {
                 return 0;
             }
