@@ -108,14 +108,14 @@ public abstract class SSLUtilBase implements SSLUtil {
         List<String> enabledProtocols =
                 getEnabled("protocols", getLog(), warnTls13, configuredProtocols, implementedProtocols);
         if (enabledProtocols.contains("SSLv3")) {
-            log.warn(sm.getString("jsse.ssl3"));
+            log.warn(sm.getString("sslUtilBase.ssl3"));
         }
         this.enabledProtocols = enabledProtocols.toArray(new String[0]);
 
         if (enabledProtocols.contains(Constants.SSL_PROTO_TLSv1_3) &&
                 sslHostConfig.getCertificateVerification() == CertificateVerification.OPTIONAL &&
                 !isTls13RenegAuthAvailable() && warnTls13) {
-            log.warn(sm.getString("jsse.tls13.auth"));
+            log.warn(sm.getString("sslUtilBase.tls13.auth"));
         }
 
         // Calculate the enabled ciphers
@@ -222,7 +222,7 @@ public abstract class SSLUtilBase implements SSLUtil {
             // Re-throw. Caller will catch and log as required
             throw ioe;
         } catch(Exception ex) {
-            String msg = sm.getString("jsse.keystore_load_failed", type, path,
+            String msg = sm.getString("sslUtilBase.keystore_load_failed", type, path,
                     ex.getMessage());
             log.error(msg, ex);
             throw new IOException(msg);
@@ -296,9 +296,19 @@ public abstract class SSLUtilBase implements SSLUtil {
 
         char[] keyPassArray = keyPass.toCharArray();
 
+        KeyManagerFactory kmf = KeyManagerFactory.getInstance(algorithm);
+        if (kmf.getProvider().getInfo().indexOf("FIPS") != -1) {
+            // FIPS doesn't like ANY wrapping nor key manipulation.
+            if (keyAlias != null) {
+                log.warn(sm.getString("sslUtilBase.aliasIgnored", keyAlias));
+            }
+            kmf.init(ksUsed, keyPassArray);
+            return kmf.getKeyManagers();
+        }
+
         if (ks == null) {
             if (certificate.getCertificateFile() == null) {
-                throw new IOException(sm.getString("jsse.noCertFile"));
+                throw new IOException(sm.getString("sslUtilBase.noCertFile"));
             }
 
             PEMFile privateKeyFile = new PEMFile(
@@ -323,11 +333,11 @@ public abstract class SSLUtilBase implements SSLUtil {
                     chain.toArray(new Certificate[0]));
         } else {
             if (keyAlias != null && !ks.isKeyEntry(keyAlias)) {
-                throw new IOException(sm.getString("jsse.alias_no_key_entry", keyAlias));
+                throw new IOException(sm.getString("sslUtilBase.alias_no_key_entry", keyAlias));
             } else if (keyAlias == null) {
                 Enumeration<String> aliases = ks.aliases();
                 if (!aliases.hasMoreElements()) {
-                    throw new IOException(sm.getString("jsse.noKeys"));
+                    throw new IOException(sm.getString("sslUtilBase.noKeys"));
                 }
                 while (aliases.hasMoreElements() && keyAlias == null) {
                     keyAlias = aliases.nextElement();
@@ -336,7 +346,7 @@ public abstract class SSLUtilBase implements SSLUtil {
                     }
                 }
                 if (keyAlias == null) {
-                    throw new IOException(sm.getString("jsse.alias_no_key_entry", (Object) null));
+                    throw new IOException(sm.getString("sslUtilBase.alias_no_key_entry", (Object) null));
                 }
             }
 
@@ -358,7 +368,6 @@ public abstract class SSLUtilBase implements SSLUtil {
         }
 
 
-        KeyManagerFactory kmf = KeyManagerFactory.getInstance(algorithm);
         kmf.init(ksUsed, keyPassArray);
 
         KeyManager[] kms = kmf.getKeyManagers();
@@ -402,7 +411,7 @@ public abstract class SSLUtilBase implements SSLUtil {
              Class<?> clazz = classLoader.loadClass(className);
              if(!(TrustManager.class.isAssignableFrom(clazz))){
                 throw new InstantiationException(sm.getString(
-                        "jsse.invalidTrustManagerClassName", className));
+                        "sslUtilBase.invalidTrustManagerClassName", className));
              }
              Object trustManagerObject = clazz.getConstructor().newInstance();
              TrustManager trustManager = (TrustManager) trustManagerObject;
@@ -429,11 +438,11 @@ public abstract class SSLUtilBase implements SSLUtil {
                 tmf.init(trustStore);
                 tms = tmf.getTrustManagers();
                 if (crlf != null && crlf.length() > 0) {
-                    throw new CRLException(sm.getString("jsseUtil.noCrlSupport", algorithm));
+                    throw new CRLException(sm.getString("sslUtilBase.noCrlSupport", algorithm));
                 }
                 // Only warn if the attribute has been explicitly configured
                 if (sslHostConfig.isCertificateVerificationDepthConfigured()) {
-                    log.warn(sm.getString("jsseUtil.noVerificationDepth", algorithm));
+                    log.warn(sm.getString("sslUtilBase.noVerificationDepth", algorithm));
                 }
             }
         }
@@ -454,7 +463,7 @@ public abstract class SSLUtilBase implements SSLUtil {
                         try {
                             ((X509Certificate) cert).checkValidity(now);
                         } catch (CertificateExpiredException | CertificateNotYetValidException e) {
-                            String msg = sm.getString("jsseUtil.trustedCertNotValid", alias,
+                            String msg = sm.getString("sslUtilBase.trustedCertNotValid", alias,
                                     ((X509Certificate) cert).getSubjectDN(), e.getMessage());
                             if (log.isDebugEnabled()) {
                                 log.debug(msg, e);
@@ -464,7 +473,7 @@ public abstract class SSLUtilBase implements SSLUtil {
                         }
                     } else {
                         if (log.isDebugEnabled()) {
-                            log.debug(sm.getString("jsseUtil.trustedCertNotChecked", alias));
+                            log.debug(sm.getString("sslUtilBase.trustedCertNotChecked", alias));
                         }
                     }
                 }
